from pywebvue.modules.VTK.core import HybridView
from pywebvue.modules.ParaView.core import apply_default_interaction_settings

from paraview import servermanager
from paraview.modules.vtkPVClientWeb import vtkPVWebApplication

from vtkmodules.web.utils import mesh as mesh_vtk

# -----------------------------------------------------------------------------
# Basic application setup
# -----------------------------------------------------------------------------

vue_use = ["vtk"]

# -----------------------------------------------------------------------------


class Helper:
    def __init__(self, app):
        self._root_protocol = None
        self._app = app
        self._pv_core = vtkPVWebApplication()
        self._pv_core.SetImageEncoding(0)
        self._hybrid_views = {}

        # Link our custom protocols initialization
        app.add_protocol_to_configure(self.configure_protocol)

    def id(self, pv_proxy):
        if pv_proxy:
            return pv_proxy.GetGlobalIDAsString()
        return ""

    def object(self, pv_id):
        try:
            pv_id = int(pv_id)
        except:
            return None
        if pv_id <= 0:
            return None
        return servermanager._getPyProxy(
            servermanager.ActiveConnection.Session.GetRemoteObject(pv_id)
        )

    def mesh(self, proxy, field_to_keep=None, point_arrays=None, cell_arrays=None):
        proxy.UpdatePipeline()
        source = proxy.GetClientSideObject()
        dataset = source.GetOutput()
        return mesh_vtk(
            dataset,
            field_to_keep=field_to_keep,
            point_arrays=point_arrays,
            cell_arrays=cell_arrays,
        )

    def scene(self, view_proxy):
        # flush data without requireing a render/picture
        tmp = view_proxy.SuppressRendering
        view_proxy.SuppressRendering = 1
        view_proxy.StillRender()
        view_proxy.SuppressRendering = tmp

        return self._app.protocol_call(
            "viewport.geometry.view.get.state", self.id(view_proxy), True
        )

    def push_image(self, view_proxy):
        if view_proxy.EnableRenderOnInteraction:
            view_proxy.EnableRenderOnInteraction = 0
        return self._app.protocol_call(
            "viewport.image.push", {"view": self.id(view_proxy)}
        )

    def camera(self, view_proxy):
        view_proxy.UpdatePropertyInformation()
        return {
            "focalPoint": list(view_proxy.CameraFocalPoint),
            "parallelProjection": view_proxy.CameraParallelProjection,
            "parallelScale": view_proxy.CameraParallelScale,
            "position": list(view_proxy.CameraPosition),
            "viewAngle": view_proxy.CameraViewAngle,
            "viewUp": list(view_proxy.CameraViewUp),
            "centerOfRotation": list(view_proxy.CenterOfRotation),
        }

    def set_camera(self, view_proxy, **kwargs):
        if "focalPoint" in kwargs:
            view_proxy.CameraFocalPoint = kwargs["focalPoint"]

        if "parallelProjection" in kwargs:
            view_proxy.CameraParallelProjection = kwargs["parallelProjection"]

        if "parallelScale" in kwargs:
            view_proxy.CameraParallelScale = kwargs["parallelScale"]

        if "position" in kwargs:
            view_proxy.CameraPosition = kwargs["position"]

        if "viewAngle" in kwargs:
            view_proxy.CameraViewAngle = kwargs["viewAngle"]

        if "viewUp" in kwargs:
            view_proxy.CameraViewUp = kwargs["viewUp"]

        if "centerOfRotation" in kwargs:
            view_proxy.CenterOfRotation = kwargs["centerOfRotation"]

    def configure_protocol(self, protocol):
        self._root_protocol = protocol

        from paraview.web.protocols import (
            ParaViewWebMouseHandler,
            ParaViewWebViewPort,
            ParaViewWebPublishImageDelivery,
            ParaViewWebLocalRendering,
        )
        from pywebvue.modules.VTK.addon_serializer import registerAddOnSerializers

        # Initialize vtk application helper
        self._root_protocol.setSharedObject("app", self._pv_core)

        # Remote rendering - image delivery
        self._root_protocol.registerLinkProtocol(ParaViewWebMouseHandler())
        self._root_protocol.registerLinkProtocol(ParaViewWebViewPort())
        self._root_protocol.registerLinkProtocol(
            ParaViewWebPublishImageDelivery(decode=False)
        )

        # Remote rendering - geometry delivery
        self._root_protocol.registerLinkProtocol(ParaViewWebLocalRendering())

        # Add custom serializer ahead of proper vtk integration
        registerAddOnSerializers()

        # Mimic client interactor on server side
        apply_default_interaction_settings()

    def add_hybrid_view(
        self,
        name,
        view,
        mode="local",
        interactive_ratio=1,
        interactive_quality=60,
        still_ratio=1,
        still_quality=98,
        **kwargs,
    ):
        if name in self._hybrid_views:
            raise ValueError(f"A view with name ({name}) is already registered")

        view_helper = HybridView(
            self,
            view,
            name,
            mode,
            interactive_ratio,
            interactive_quality,
            still_ratio,
            still_quality,
        )
        self._hybrid_views[name] = view_helper
        return view_helper


# -----------------------------------------------------------------------------
# Module advanced initialization
# -----------------------------------------------------------------------------

HELPER = None


def setup(app, **kwargs):
    global HELPER
    HELPER = Helper(app)


# -----------------------------------------------------------------------------
# Helper methods only valid once the module has been enabled
# -----------------------------------------------------------------------------


def id(vtk_obj):
    return HELPER.id(vtk_obj)


def object(vtk_id):
    return HELPER.object(vtk_id)


def mesh(dataset, field_to_keep=None, point_arrays=None, cell_arrays=None):
    return HELPER.mesh(dataset, field_to_keep, point_arrays, cell_arrays)


def scene(render_window):
    return HELPER.scene(render_window)


def push_image(render_window):
    return HELPER.push_image(render_window)


def camera(render_window):
    return HELPER.camera(render_window)


def set_camera(render_window, **kwargs):
    return HELPER.set_camera(render_window, **kwargs)


def view(view, name="view", **kwargs):
    return HELPER.add_hybrid_view(name, view, **kwargs)



import os
import json

serve_path = os.path.abspath(os.path.join(os.path.dirname(__file__), "serve"))

serve = {"__markdown_it": serve_path}
scripts = ["/__markdown_it/markdown-it-vue.umd.min.js"]
styles = ["/__markdown_it/markdown-it-vue.css"]



import json

from vtkmodules.web import getReferenceId
from vtkmodules.web.render_window_serializer import (
    wrapId,
    registerInstanceSerializer,
    serializeInstance,
)


def cubeAxesSerializer(parent, actor, actorId, context, depth):
    """
    Possible add-on properties for vtk.js:
        gridLines: True,
        axisLabels: None,
        axisTitlePixelOffset: 35.0,
        axisTextStyle: {
            fontColor: 'white',
            fontStyle: 'normal',
            fontSize: 18,
            fontFamily: 'serif',
        },
        tickLabelPixelOffset: 12.0,
        tickTextStyle: {
            fontColor: 'white',
            fontStyle: 'normal',
            fontSize: 14,
            fontFamily: 'serif',
        },
    """
    axisLabels = ["", "", ""]
    if actor.GetXAxisLabelVisibility():
        axisLabels[0] = actor.GetXTitle()
    if actor.GetYAxisLabelVisibility():
        axisLabels[1] = actor.GetYTitle()
    if actor.GetZAxisLabelVisibility():
        axisLabels[2] = actor.GetZTitle()

    return {
        "parent": getReferenceId(parent),
        "id": actorId,
        "type": "vtkCubeAxesActor",
        "properties": {
            # vtkProp
            "visibility": actor.GetVisibility(),
            "pickable": actor.GetPickable(),
            "dragable": actor.GetDragable(),
            "useBounds": actor.GetUseBounds(),
            # vtkProp3D
            "origin": actor.GetOrigin(),
            "position": actor.GetPosition(),
            "scale": actor.GetScale(),
            # vtkActor
            "forceOpaque": actor.GetForceOpaque(),
            "forceTranslucent": actor.GetForceTranslucent(),
            # vtkCubeAxesActor
            "dataBounds": actor.GetBounds(),
            "faceVisibilityAngle": 8,
            "gridLines": True,
            "axisLabels": axisLabels,
            "axisTitlePixelOffset": 35.0,
            "axisTextStyle": {
                "fontColor": "white",
                "fontStyle": "normal",
                "fontSize": 18,
                "fontFamily": "serif",
            },
            "tickLabelPixelOffset": 12.0,
            "tickTextStyle": {
                "fontColor": "white",
                "fontStyle": "normal",
                "fontSize": 14,
                "fontFamily": "serif",
            },
        },
        "calls": [["setCamera", [wrapId(getReferenceId(actor.GetCamera()))]]],
        "dependencies": [],
    }


def scalarBarActorSerializer(parent, actor, actorId, context, depth):
    dependencies = []
    calls = []
    lut = actor.GetLookupTable()
    if not lut:
        return None

    lutId = getReferenceId(lut)
    lutInstance = serializeInstance(actor, lut, lutId, context, depth + 1)
    if not lutInstance:
        return None

    dependencies.append(lutInstance)
    calls.append(["setScalarsToColors", [wrapId(lutId)]])

    prop = None
    if hasattr(actor, "GetProperty"):
        prop = actor.GetProperty()
    else:
        if context.debugAll:
            print("This scalarBarActor does not have a GetProperty method")

        if prop:
            propId = getReferenceId(prop)
            propertyInstance = serializeInstance(
                actor, prop, propId, context, depth + 1
            )
            if propertyInstance:
                dependencies.append(propertyInstance)
                calls.append(["setProperty", [wrapId(propId)]])

    axisLabel = actor.GetTitle()
    width = actor.GetWidth()
    height = actor.GetHeight()

    # axisTitlePixelOffset = actor.GetTextPad()
    # position = actor.GetPosition()
    # position2 = actor.GetPosition2()
    # print(f'axisTitlePixelOffset: {axisTitlePixelOffset}')
    # print(f'axisLabel: {axisLabel}')
    # print(f'width: {width}')
    # print(f'height: {height}')
    # print(f'position: {position}')
    # print(f'position2: {position2}')

    return {
        "parent": getReferenceId(parent),
        "id": actorId,
        "type": "vtkScalarBarActor",
        "properties": {
            # vtkProp
            "visibility": actor.GetVisibility(),
            "pickable": actor.GetPickable(),
            "dragable": actor.GetDragable(),
            "useBounds": actor.GetUseBounds(),
            # vtkActor2D
            # "position": actor.GetPosition(),
            # "position2": actor.GetPosition2(),
            # "width": actor.GetWidth(),
            # "height": actor.GetHeight(),
            # vtkScalarBarActor
            "automated": True,
            "axisLabel": axisLabel,
            # 'barPosition': [0, 0],
            # 'barSize': [0, 0],
            "boxPosition": [0.88, -0.92],
            "boxSize": [width, height],
            "axisTitlePixelOffset": 36.0,
            "axisTextStyle": {
                "fontColor": "white",
                "fontStyle": "normal",
                "fontSize": 18,
                "fontFamily": "serif",
            },
            "tickLabelPixelOffset": 14.0,
            "tickTextStyle": {
                "fontColor": "white",
                "fontStyle": "normal",
                "fontSize": 14,
                "fontFamily": "serif",
            },
        },
        "calls": calls,
        "dependencies": dependencies,
    }


def registerAddOnSerializers():
    registerInstanceSerializer("vtkCubeAxesActor", cubeAxesSerializer)
    registerInstanceSerializer("vtkScalarBarActor", scalarBarActorSerializer)



import os
import base64
import mimetypes
import time
import threading
import inspect
import asyncio

mimetypes.init()

MONITORS = {}


def base_directory():
    frame = inspect.stack()[2]
    module = inspect.getmodule(frame[0])
    if module is None:
        return os.getcwd()
    return os.path.abspath(os.path.dirname(module.__file__))


def abs_path(file_path, root=None):
    basepath = os.getcwd()
    if root:
        if os.path.isfile(root):
            basepath = os.path.dirname(root)
        elif os.path.isdir(root):
            basepath = root

    full_path = file_path
    if not os.path.isabs(full_path):
        full_path = os.path.join(basepath, file_path)

    return os.path.abspath(full_path)


def read_file_as_txt(file_path, root=None):
    full_path = abs_path(file_path, root)
    with open(full_path) as txt_file:
        return str(txt_file.read())


def to_mime(file_path):
    return mimetypes.guess_type(file_path)[0]


def read_file_as_base64_url(file_path, root=None):
    full_path = abs_path(file_path, root)
    with open(full_path, "rb") as bin_file:
        encoded = base64.b64encode(bin_file.read()).decode("ascii")
        mime = to_mime(file_path)
        return f"data:{mime};base64,{encoded}"


def validate(txt, method, root):
    if not txt:
        return False

    if len(txt) < 256:
        full_path = abs_path(txt, root)
        if os.path.isfile(full_path):
            return method(full_path)

    return txt


def monitor(key, txt, method, root):
    global MONITORS
    loop = asyncio.get_event_loop()
    full_path = abs_path(txt, root)
    if os.path.isfile(full_path):
        if key in MONITORS:
            if MONITORS[key].get("full_path") == full_path:
                return
            MONITORS[key]["running"] = False
            MONITORS[key].get("thread").join()

        MONITORS[key] = {
            "full_path": full_path,
            "running": True,
        }

        def update(loop):
            last_ts = os.stat(full_path).st_mtime
            while MONITORS[key]["running"]:
                time.sleep(1)
                ts = os.stat(full_path).st_mtime
                if ts != last_ts:
                    last_ts = ts
                    loop.call_soon_threadsafe(method, read_file_as_txt(full_path))

        thread = threading.Thread(target=update, args=(loop,))
        thread.daemon = True
        thread.start()
        MONITORS[key]["thread"] = thread


def stop_all_monitors():
    global MONITORS
    for key in MONITORS:
        MONITORS[key]["running"] = False
        MONITORS[key]["thread"].join()


def clean_state(state):
    cleaned = {}
    for key in state:
        cleaned[key] = clean_value(state[key])

    return cleaned


def clean_value(value):
    if isinstance(value, dict) and "_filter" in value.keys():
        subset = {}
        subset.update(value)
        keys_to_filter = value.get("_filter")
        for key in keys_to_filter:
            subset.pop(key, None)
        return subset

    if isinstance(value, list):
        return list(map(clean_value, value))

    return value


# =============================================================================
# Internal classes
# =============================================================================


class CallbackSet:
    def __init__(self):
        self.data = []

    def add(self, callback):
        if callable(callback):
            found = False
            for c in self.data:
                if c == callback:
                    return
            self.data.append(callback)
        else:
            for c in callback:
                self.add(c)

    def clear(self):
        self.data = []

    def __iter__(self):
        return iter(self.data)


# =============================================================================
# ChangeHandler
# =============================================================================


class ChangeHandler:
    def __init__(self, app, known_state={}):
        self._app = app
        self._protocol = app.protocol
        self._known_state = known_state
        self._modified_keys = set()
        self._all_modified_keys = set()
        self._actions = []

    def modified(self, name, value):
        self._modified_keys.add(name)

    def add_action(self, ref, args=[], value=None, method=None, property=None):
        action = {"ref": ref}
        if method:
            action["type"] = "method"
            action["method"] = method
            action["args"] = args

        if property:
            action["type"] = "property"
            action["property"] = property
            action["value"] = value

        self._actions.append(action)

    def clear_actions(self):
        self._actions.clear()

    @property
    def modified_state(self):
        modified_state = {}
        for key in self._all_modified_keys:
            if (
                key in self._known_state
                and self._known_state[key] == self._app.state[key]
            ):
                pass
            else:
                modified_state[key] = clean_value(self._app.state[key])

        return modified_state

    @property
    def actions(self):
        return self._actions

    def __enter__(self):
        self._app._dirty_set = self._known_state
        if self not in self._app._change_handlers:
            self._app._change_handlers.append(self)
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        callbacks = CallbackSet()
        while len(self._modified_keys):
            # Gather callbacks
            for key in self._modified_keys:
                if key in self._app._change_callbacks:
                    callbacks.add(self._app._change_callbacks[key])

            self._all_modified_keys.update(self._modified_keys)
            self._modified_keys.clear()

            # Process callbacks
            for fn in callbacks:
                try:
                    # Pass state as **kwargs
                    fn(**self._app.state)
                except TypeError:
                    fn()

            callbacks.clear()

        self._app._change_handlers.remove(self)
        self._app._dirty_set = None

        if self._protocol:
            self._protocol.push_state_change(self.modified_state)

            if len(self._actions):
                self._protocol.push_actions(self._actions)



import os
import json

serve_path = os.path.abspath(os.path.join(os.path.dirname(__file__), "serve"))

serve = {"__vue_deck": serve_path}
scripts = ["/__vue_deck/vue-deck.umd.min.js"]
vue_use = ["VueDeck"]


def to_jsonInput(deck):
    return json.loads(deck.to_json())



import os
import sys
import shutil
import json
import asyncio

from pywebvue.utils import abs_path

from wslink import server, launcher
from wslink import register as exportRpc

from wslink.websocket import ServerProtocol
from wslink.websocket import LinkProtocol

# Only used for debug
import json

WWW_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "www")


def current_cmd(
    skip_launcher=True, skip_port=True, add_on=[], default_skip=["--name", "--deploy"]
):
    cmd = ["python"]
    skip_next = False
    for item in sys.argv:
        if skip_next:
            skip_next = False
            continue

        if skip_port and item == "--port":
            skip_next = True
            continue

        if item in default_skip:
            skip_next = True
            continue

        if skip_launcher and item == "--launcher":
            continue

        if len(cmd) == 1:
            cmd.append(os.path.abspath(item))
        else:
            cmd.append(item)

    for item in add_on:
        cmd.append(item)

    return cmd


DEFAULT_LAUNCHER_CONFIG = {
    "configuration": {
        "host": "0.0.0.0",
        "port": 8080,
        "endpoint": "paraview",
        "log_dir": "/tmp",
        "proxy_file": "/tmp/pywebproxy.txt",
        "sessionURL": "ws://USE_HOSTNAME:${port}/ws",
        "timeout": 25,
        "content": WWW_DIR,
        "sanitize": {},
        "fields": [],
    },
    "resources": [{"host": "localhost", "port_range": [9001, 9100]}],
    "properties": {},
    "apps": {
        "PyWebVue": {
            "cmd": current_cmd(add_on=["--port", "$port"]),
            "ready_line": "Starting factory",
        },
    },
}


class CoreServer(ServerProtocol):
    authentication_token = "wslink-secret"
    app = None
    DEBUG = False

    # ---------------------------------------------------------------
    # Static methods
    # ---------------------------------------------------------------

    @staticmethod
    def add_arguments(parser):
        server.add_arguments(parser)

    @staticmethod
    def configure(args):
        CoreServer.authentication_token = args.authKey

    @staticmethod
    def start_webserver(args):
        server.start_webserver(options=args, protocol=CoreServer)

    @staticmethod
    def get_launcher_config(args):
        config = {}
        config.update(DEFAULT_LAUNCHER_CONFIG)

        config["configuration"]["port"] = args.port

        if "configuration" in CoreServer.app.launcher:
            config["configuration"].update(CoreServer.app.launcher["configuration"])

        if "resources" in CoreServer.app.launcher:
            config["resources"] = CoreServer.app.launcher["resources"]

        if "apps" in CoreServer.app.launcher:
            config["apps"].update(CoreServer.app.launcher["apps"])

        return config

    @staticmethod
    def start_launcher(args):
        config = CoreServer.get_launcher_config(args)

        if CoreServer.DEBUG:
            print(json.dumps(config, indent=2))

        print("Starting launcher...")
        launcher.startWebServer(args, config)

    @staticmethod
    def deploy_setup(args):
        work_dir = os.path.abspath(args.deploy)
        out_www = os.path.join(work_dir, "www")
        os.makedirs(work_dir, exist_ok=True)
        for directory in ["www", "apps", "launcher", "logs"]:
            os.makedirs(os.path.join(work_dir, directory), exist_ok=True)

        # copy app
        app_name = args.name
        app_dst_dir = os.path.join(work_dir, "apps", app_name)
        app_src_dir = abs_path("./", CoreServer.app._root)
        shutil.copytree(app_src_dir, app_dst_dir, dirs_exist_ok=True)

        # copy www
        shutil.copytree(WWW_DIR, out_www, dirs_exist_ok=True)
        for sub_dir in CoreServer.app.serve:
            src_dir = abs_path(CoreServer.app.serve[sub_dir], CoreServer.app._root)
            dst_dir = os.path.join(out_www, sub_dir)
            shutil.copytree(src_dir, dst_dir, dirs_exist_ok=True)

        # update launcher config
        launcher_conf = os.path.join(work_dir, "launcher", "config.json")
        config = CoreServer.get_launcher_config(args)
        config["configuration"]["log_dir"] = "${work_dir}/logs"
        config["configuration"]["proxy_file"] = "${work_dir}/launcher/proxy.txt"
        config["configuration"]["sessionURL"] = "ws://USE_HOSTNAME:${port}/ws"
        config["configuration"]["content"] = "${work_dir}/www"
        config["properties"]["work_dir"] = work_dir

        try:
            if os.path.exists(launcher_conf):
                with open(launcher_conf) as f:
                    conf = json.load(f)
                    apps = conf.get("apps", {})
                    for name in apps:
                        config["apps"][name] = apps[name]
        except:
            pass

        # Override our app
        config["apps"][app_name] = {
            "cmd": current_cmd(add_on=["--port", "$port"]),
            "ready_line": "Starting factory",
        }
        config["apps"][app_name]["cmd"][1] = config["apps"][app_name]["cmd"][1].replace(
            app_src_dir, "${work_dir}/" + app_name
        )

        # Remove any invalid app
        apps = config.get("apps", {})
        to_delete = []
        for name in apps:
            app_path = apps[name]["cmd"][1]
            if not app_path.startswith("${work_dir}"):
                to_delete.append(name)

        for name in to_delete:
            apps.pop(name)

        with open(launcher_conf, "w") as json_file:
            json.dump(config, json_file, indent=2)

    # ---------------------------------------------------------------
    # Server
    # ---------------------------------------------------------------

    def initialize(self):  # Called by wslink
        self.rpcMethods = {}
        self.app = CoreServer.app
        self.app._root_protocol = self

        for configure in self.app._protocols_to_configure:
            configure(self)

        self.updateSecret(CoreServer.authentication_token)

        if self.app.on_ready:
            self.app.on_ready()

    # ---------------------------------------------------------------

    def getRPCMethod(self, name):
        if len(self.rpcMethods) == 0:
            import inspect

            for protocolObject in self.getLinkProtocols():
                test = lambda x: inspect.ismethod(x) or inspect.isfunction(x)
                for k in inspect.getmembers(protocolObject.__class__, test):
                    proc = k[1]
                    if "_wslinkuris" in proc.__dict__:
                        uri_info = proc.__dict__["_wslinkuris"][0]
                        if "uri" in uri_info:
                            uri = uri_info["uri"]
                            self.rpcMethods[uri] = (protocolObject, proc)

        if name in self.rpcMethods:
            return self.rpcMethods[name]

    # ---------------------------------------------------------------
    # Publish
    # ---------------------------------------------------------------

    def push_state_change(self, modified_state):
        if CoreServer.DEBUG:
            print(f"=> Push state {len(json.dumps(modified_state))}")
        self.publish("topic.ws.vue.state", modified_state)

    # ---------------------------------------------------------------

    def push_actions(self, actions):
        self.publish("topic.ws.vue.actions", actions)

    # ---------------------------------------------------------------

    def push_layout(self, layout):
        # publish might not be bind if no client is connected
        # in debug layout will be pushed when changed automatically
        if hasattr(self, "publish"):
            self.publish("topic.ws.vue.layout", layout)

    # ---------------------------------------------------------------

    def push_routes(self, routes):
        self.publish("topic.ws.vue.routes", routes)

    # ---------------------------------------------------------------
    # RPCs
    # ---------------------------------------------------------------

    @exportRpc("ws.vue.init")
    def get_state(self):
        return self.app.get_initial_state()

    # ---------------------------------------------------------------

    @exportRpc("ws.vue.trigger")
    async def trigger(self, name, args, kwargs):
        with self.app.capture_changes():
            if name in self.app._triggers:
                await asyncio.coroutine(self.app._triggers[name])(*args, **kwargs)
            else:
                print(f"Tigger {name} seems to be missing")

    # ---------------------------------------------------------------

    @exportRpc("ws.vue.state.update")
    def update_state(self, changes):
        values_to_ignore = {}
        for change in changes:
            values_to_ignore[change["key"]] = (
                change["value"] if "value" in change else None
            )

        with self.app.capture_changes(values_to_ignore):
            for change in changes:
                self.app.set(
                    change["key"], change["value"] if "value" in change else None
                )



import os
import argparse
from .protocol import CoreServer
from .utils import (
    ChangeHandler,
    read_file_as_txt,
    read_file_as_base64_url,
    validate,
    abs_path,
    monitor,
    clean_state,
    clean_value,
    base_directory,
)

from wslink import server

WWW_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "www")


class App:
    def __init__(
        self,
        name,
        debug=False,
        eager_state_listening=True,
        **kwargs,
    ):
        self.name = name
        self._root = base_directory()
        self._debug = debug
        self._eager_state_listening = eager_state_listening

        self.on_ready = None
        self.active_objects = {}

        # May need to protect but aimed to be set by user
        self.vuetify = {}
        self.state = {}
        self.scripts = []
        self.styles = []
        self.vue_use = ["vuetify"]
        self.serve = {}
        self.launcher = {}

        # properties
        self._favicon = None
        self._layout = "<div>Ws Vue default</div>"
        self._layout_monitor = None
        self._routes = []

        # @change
        self._dirty_set = None
        self._change_callbacks = {}
        self._change_handlers = []
        self._server_keys = set()
        self._client_keys = set()

        # @trigger
        self._triggers = {}

        # CLI argument handling
        self._parser = None

        # modules
        self._loaded_modules = set()

        # protocols to register
        self._protocols_to_configure = []
        self._root_protocol = None

        # Use ./template.html if available
        if os.path.exists(abs_path("./template.html", self._root)):
            self.layout = "./template.html"

    def __getitem__(self, item):
        return getattr(self, item)

    # -------------------------------------------------------------------------
    # State management helpers
    # -------------------------------------------------------------------------

    def listen(self, *_args):
        """
        This method needs to be call explicitly if the server expect to use
        a state variable that can be updated by the client but that is not
        listed inside any @app.change(). This allow the server to properly
        describe to the client its dependencies at initialization time.
        This usually happen when app.get() is called inside methods and
        does not need to leverage the @app.change() callback on itself.
        """
        for name in _args:
            self._server_keys.add(name)

    def client_only(self, *_args):
        for name in _args:
            self._client_keys.add(name)

    def flush_state(self, *_args):
        state_to_flush = {}
        if len(_args):
            for key in _args:
                state_to_flush[key] = clean_value(self.state[key])
        else:
            # flush any pending ChangeHandler
            for change_handler in self._change_handlers:
                state_to_flush.update(change_handler.modified_state)

        if self.protocol:
            self.protocol.push_state_change(state_to_flush)

    def flush_actions(self):
        actions_to_flush = []
        for change_handler in self._change_handlers:
            actions_to_flush.extend(change_handler.actions)
            change_handler.clear_actions()

        if self.protocol and len(actions_to_flush):
            self.protocol.push_actions(actions_to_flush)

    # -------------------------------------------------------------------------
    # Initialization helper
    # -------------------------------------------------------------------------

    def enableModule(self, module, **kwargs):
        if module in self._loaded_modules:
            return

        if "setup" in module.__dict__:
            module.setup(self, **kwargs)

        for key in ["scripts", "styles", "vue_use"]:
            if key in module.__dict__:
                setattr(self, key, self[key] + module.__dict__[key])

        for key in ["state", "serve", "vuetify"]:
            if key in module.__dict__:
                self[key].update(module.__dict__[key])

        self._loaded_modules.add(module)

    def disableModule(self, module):
        for key in ["scripts", "styles", "vue_use"]:
            if key in module.__dict__:
                for item in module.__dict__[key]:
                    self[key].remove(item)

        for key in ["state", "serve", "vuetify"]:
            if key in module.__dict__:
                for entry in module.__dict__[key]:
                    del self[key][entry]

    # -------------------------------------------------------------------------
    # Annotations
    # -------------------------------------------------------------------------

    def change(self, *_args, **_kwargs):
        """
        Use as decorator `@app.change(key1, key2, ...)`
        """

        def register_change_callback(func):
            for name in _args:
                if name not in self._change_callbacks:
                    self._change_callbacks[name] = []
                self._change_callbacks[name].append(func)
                self._server_keys.add(name)
            return func

        return register_change_callback

    # -------------------------------------------------------------------------

    def trigger(self, name):
        """
        Use as decorator `@app.trigger(name)`
        """

        def register_trigger(func):
            self._triggers[name] = func
            return func

        return register_trigger

    # -------------------------------------------------------------------------
    # CLI handling
    # -------------------------------------------------------------------------

    @property
    def cli_parser(self):
        if self._parser:
            return self._parser

        self._parser = argparse.ArgumentParser(description="PyWebVue server By Kitware")

        # Launcher / multi-users
        self._parser.add_argument(
            "--launcher",
            help="Start launcher process rather than single process server",
            action="store_true",
        )

        # Deploy helper
        self._parser.add_argument("--deploy", help="Prepare a deployable directory")
        self._parser.add_argument(
            "--name",
            default="pywebvue",
            help="Name to use for that application in the deploy tree",
        )

        CoreServer.add_arguments(self._parser)

        return self._parser

    @property
    def cli_args(self):
        return self.cli_parser.parse_known_args()[0]

    # -------------------------------------------------------------------------
    # Properties
    # -------------------------------------------------------------------------

    @property
    def layout(self):
        return self._layout

    @layout.setter
    def layout(self, value):
        self._layout = validate(value, read_file_as_txt, self._root)
        if self.protocol:
            self.protocol.push_layout(self._layout)

        if self._debug:
            monitor("layout", value, lambda v: setattr(self, "layout", v), self._root)

    # -------------------------------------------------------------------------

    @property
    def favicon(self):
        return self._favicon

    @favicon.setter
    def favicon(self, value):
        self._favicon = validate(value, read_file_as_base64_url, self._root)

    # -------------------------------------------------------------------------

    @property
    def routes(self):
        return self._routes

    @routes.setter
    def routes(self, value):
        self._routes = value
        if self.protocol:
            self.protocol.push_routes(self._routes)

    # -------------------------------------------------------------------------
    # API client / server
    # -------------------------------------------------------------------------

    def get(self, key):
        return self.state[key]

    # -------------------------------------------------------------------------

    def set(self, key, value, force=False):
        if (
            key not in self.state
            or self.state[key] != value
            or force
            or key not in self._server_keys
        ):
            self.state[key] = value
            for change_handler in self._change_handlers:
                change_handler.modified(key, value)

    # -------------------------------------------------------------------------

    def dirty(self, *args):
        if self._dirty_set is not None:
            return len(set(args).intersection(self._dirty_set)) > 0
        return False

    # -------------------------------------------------------------------------

    def dirty_all(self, *args):
        if self._dirty_set is not None:
            return len(set(args).intersection(self._dirty_set)) == len(args)
        return False

    # -------------------------------------------------------------------------

    def update(self, **kwargs):
        for change_handler in self._change_handlers:
            change_handler.add_action(**kwargs)

    # -------------------------------------------------------------------------

    def capture_changes(self, known_state={}):
        return ChangeHandler(self, known_state)

    # -------------------------------------------------------------------------
    # Helper methods
    # -------------------------------------------------------------------------

    def url(self, file_path):
        return read_file_as_base64_url(file_path, self._root)

    # -------------------------------------------------------------------------

    def txt(self, file_path):
        return read_file_as_txt(file_path, self._root)

    # -------------------------------------------------------------------------

    # -------------------------------------------------------------------------
    # common
    # -------------------------------------------------------------------------

    def get_initial_state(self):
        state = {
            "name": self.name,
            "vuetify": self.vuetify,
            "layout": self.layout,
            "favicon": self.favicon,
            "routes": self.routes,
            "state": clean_state(self.state),
            "scripts": self.scripts,
            "use": self.vue_use,
            "styles": self.styles,
            "stateListening": list(self._server_keys - self._client_keys),
        }
        return state

    # -------------------------------------------------------------------------

    def add_protocol_to_configure(self, configure_protocol_fn):
        self._protocols_to_configure.append(configure_protocol_fn)

    @property
    def protocol(self):
        return self._root_protocol

    # -------------------------------------------------------------------------

    def protocol_call(self, method, *args, **kwargs):
        if self.protocol:
            pair = self.protocol.getRPCMethod(method)
            if pair:
                obj, func = pair
                return func(obj, *args, **kwargs)

    # -------------------------------------------------------------------------
    # Start server
    # -------------------------------------------------------------------------

    def run_server(self, port=None):
        # Prevent faulty state synchronization with naive usage
        if self._eager_state_listening:
            self.listen(*list(self.state.keys()))

        CoreServer.app = self
        CoreServer.DEBUG = self._debug

        args = self.cli_args

        if port:
            args.port = port

        if not args.content:
            args.content = WWW_DIR

        if args.launcher:
            CoreServer.start_launcher(args)
            # FIXME we don't support yet self.serve with launcher...
        elif args.deploy:
            CoreServer.deploy_setup(args)
        else:
            if len(self.serve):
                endpoints = []
                for key in self.serve:
                    endpoints.append(f"{key}={abs_path(self.serve[key], self._root)}")
                args.fsEndpoints = "|".join(endpoints)
                print(args.fsEndpoints)

            CoreServer.configure(args)
            CoreServer.start_webserver(args)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import re

from setuptools import setup, find_packages

# perform the install
setup(
    name="pywebvue",
    description="Python server creating interactive vue app over WebSocket",
    long_description="Build a 3D Web application using VTK or ParaView while only writing a Python script. This project aims to streamline the existing ParaViewWeb infrastructure into something that is simpler to use by lowering the entry point complexity.",
    author="Kitware, Inc.",
    author_email="kitware@kitware.com",
    url=f"https://github.com/kitware/py-web-vue",
    license="BSD-3-Clause",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: JavaScript",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    keywords="Python Vue.js websocket VTK ParaView Web3D",
    packages=find_packages("src", exclude=("tests.*", "tests")),
    package_dir={"": "src"},
    package_data={"pywebvue": ["www/*", "www/*/*"]},
    install_requires=["wslink>=1.0.7"],
)



import os
from collections import defaultdict
import numpy as np
import pandas as pd
import simplejson
from numbers import Number
from enum import IntEnum, IntFlag

# -----------------------------------------------------------------------------

serve_path = os.path.abspath(os.path.join(os.path.dirname(__file__), "serve"))

serve = {"__ag_grid": serve_path}
scripts = ["__ag_grid/vue-ag-grid.umd.min.js"]
styles = ["__ag_grid/vue-ag-grid.css"]
vue_use = ["AgGrid"]

# -----------------------------------------------------------------------------
# numpy types: 'biufcmMOSUV' https://numpy.org/doc/stable/reference/generated/numpy.dtype.kind.html
# -----------------------------------------------------------------------------

type_mapper = {
    "b": ["textColumn"],
    "i": [],  # ["numericColumn", "numberColumnFilter"],
    "u": [],  # ["numericColumn", "numberColumnFilter"],
    "f": [],  # ["numericColumn", "numberColumnFilter"],
    "c": [],
    "m": [],  # ['timedeltaFormat'],
    "M": [],  # ["dateColumnFilter", "shortDateTimeFormat"],
    "O": [],
    "S": [],
    "U": [],
    "V": [],
}


def cast_to_serializable(value):
    isoformat = getattr(value, "isoformat", None)
    if (isoformat) and callable(isoformat):
        return isoformat()
    elif isinstance(value, Number):
        if np.isnan(value) or np.isinf(value):
            return value.__str__()
        return value

    return value.__str__()


# -----------------------------------------------------------------------------


def to_grid(
    dataframe, sortable=[], filter=[], rowGroup=[], resizable=[], headerNames={}
):
    # Setup columns definitions
    columnDefs = []
    for col_name, col_type in zip(dataframe.columns, dataframe.dtypes):
        columnDefs.append(
            {
                "headerName": headerNames[col_name]
                if col_name in headerNames
                else col_name,
                "field": col_name,
                "type": type_mapper.get(col_type.kind, []),
                "sortable": col_name in sortable,
                "filter": col_name in filter,
                # 'rowGroup': col_name in rowGroup,
                "resizable": col_name in resizable,
                # "minWidth": min_column_width,
                # "editable": editable,
            }
        )

    # Extract row data
    rowData = dataframe.applymap(cast_to_serializable).to_dict(orient="records")

    return {
        "columnDefs": columnDefs,
        "rowData": rowData,
    }


# -----------------------------------------------------------------------------


def update_columnDefs(grid, change_set={}):
    list_to_update = grid
    if "columnDefs" in grid:
        list_to_update = grid["columnDefs"]

    for item in list_to_update:
        if item["field"] in change_set:
            item.update(change_set[item["field"]])



import os

serve_path = os.path.abspath(os.path.join(os.path.dirname(__file__), "serve"))

serve = {"__vega_embed": serve_path}
scripts = ["__vega_embed/vue-vega-embed.umd.min.js"]
vue_use = ["VueVegaEmbed"]


def altair_to_spec(chart):
    return chart.to_dict()
